import argparse
from typing import Union

import core.sheetwork as upload_task
import core.task.init as init_task
from core._version import __version__
from core.config.config import ConfigLoader
from core.config.profile import Profile
from core.config.project import Project
from core.flags import FlagParser
from core.logger import log_manager
from core.utils import check_and_compare_version

parser = argparse.ArgumentParser(
    prog="sheetwork",
    formatter_class=argparse.RawTextHelpFormatter,
    description="CLI tool to load google sheets onto a DB.",
    epilog="Select one of these sub-commands to find specific help for those.",
)
parser.add_argument("-v", "--version", action="version", version=f"%(prog)s Running v{__version__}")

base_subparser = argparse.ArgumentParser(add_help=False)
base_subparser.add_argument("--log-level", help="sets the log level", type=str, default=str())
base_subparser.add_argument(
    "--profile-dir",
    help="Unusual path to the directory in which the 'profiles.yml' can be found",
    default=str(),
)
base_subparser.add_argument(
    "--project-dir",
    help="Unusual path to the directory in which the 'sheetwork_project.yml' can be found",
    default=str(),
)

# Adds sub task parsers
subs = parser.add_subparsers(title="Available sub commands", dest="command")

# Upload task parser
upload_sub = subs.add_parser(
    "upload", parents=[base_subparser], help="Pull, sanitize and upload a google sheet."
)
upload_sub.set_defaults(cls=upload_task.SheetBag, which="upload")
upload_sub.add_argument("--schema", help="Target Schema Name", type=str, default=None)
upload_sub.add_argument("--table", help="Target Table Name", type=str, default=None)
upload_sub.add_argument(
    "-sn", "--sheet-name", help="Name of your sheet from config", type=str, default=None
)
upload_sub.add_argument("-sk", "--sheet-key", help="Google sheet Key", type=str, default=None)
upload_sub.add_argument(
    "--dry-run", help="Skips pushing to database", action="store_true", default=False
)
upload_sub.add_argument(
    "-i",
    "--interactive",
    help="Turns on interactive mode, which allows previews and cleanup choices",
    action="store_true",
    default=False,
)
upload_sub.add_argument(
    "-t",
    "--target",
    help="Specity target profile. When none provided sheetwork will use the profile default",
)
upload_sub.add_argument(
    "--sheet-config-dir",
    help="Unusual path to the directory in which the 'sheets.yml' can be found",
    default=str(),
)
upload_sub.add_argument(
    "--create-table",
    help="Creates target table before pushing.",
    action="store_true",
    default=False,
)

# Init task parser
init_sub = subs.add_parser(
    "init", parents=[base_subparser], help="Initialise your sheetwork project"
)
init_sub.set_defaults(cls=init_task.InitTask, which="init")
init_sub.add_argument("--project-name", help="Name you want to init your dbt project with")
init_sub.add_argument(
    "--force-credentials-folders",
    action="store_true",
    default=False,
    help="Forces init task to at least attempt to create credential folders.",
)


def handle(parser: argparse.ArgumentParser):
    flag_parser = FlagParser(parser)
    flag_parser.consume_cli_arguments()

    if flag_parser.log_level == "debug":
        log_manager.set_debug()

    if flag_parser.args.command == "init":
        task: Union[init_task.InitTask, upload_task.SheetBag] = init_task.InitTask(flag_parser)
        return task.run()

    if flag_parser.args.command == "upload":
        project = Project(flag_parser)
        config = ConfigLoader(flag_parser, project)
        profile = Profile(project)
        task = upload_task.SheetBag(config, flag_parser, profile)
        return task.run()


def main(parser: argparse.ArgumentParser = parser):
    print(f"Sheetwork version: {__version__} \n")
    check_and_compare_version()
    if parser:
        handle(parser)
    else:
        raise NotImplementedError(
            """You did not parse any args to sheetwork.
            If you are not sure how to use it consult the help by doing: sheetwork --help
            """
        )


if __name__ == "__main__":
    main()



import os

import mock
import pytest

from core.config.config import ConfigLoader
from core.config.profile import Profile
from core.config.project import Project
from core.flags import FlagParser
from tests.mockers import (
    DIRTY_DF,
    DROP_COL_DF,
    EXCLUDED_DF_COLS,
    RENAMED_COLS,
    RENAMED_DF,
    generate_test_df,
)

FIXTURE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)))


@pytest.mark.datafiles(FIXTURE_DIR)
def test_rename_columns(datafiles):
    from core.sheetwork import SheetBag
    from core.main import parser

    flags = FlagParser(
        parser,
        test_sheet_name="df_renamer",
        project_dir=str(datafiles),
        sheet_config_dir=str(datafiles),
        profile_dir=str(datafiles),
    )
    project = Project(flags, "sheetwork_test")
    profile = Profile(project)
    config = ConfigLoader(flags, project)
    df = generate_test_df(DIRTY_DF)
    renamed_df = SheetBag(config, flags, profile).rename_columns(df)

    assert renamed_df.columns.tolist() == RENAMED_COLS


@pytest.mark.datafiles(FIXTURE_DIR)
def test_exclude_columns(datafiles):
    from core.sheetwork import SheetBag
    from core.main import parser

    flags = FlagParser(
        parser,
        test_sheet_name="df_dropper",
        project_dir=str(datafiles),
        sheet_config_dir=str(datafiles),
        profile_dir=str(datafiles),
    )
    project = Project(flags, "sheetwork_test")
    profile = Profile(project)
    config = ConfigLoader(flags, project)
    df = generate_test_df(DROP_COL_DF)
    excluded_df = SheetBag(config, flags, profile).exclude_columns(df)

    assert excluded_df.columns.tolist() == EXCLUDED_DF_COLS


@pytest.mark.datafiles(FIXTURE_DIR)
def test_load_sheet(datafiles):
    from core.sheetwork import SheetBag
    from core.main import parser

    flags = FlagParser(
        parser,
        test_sheet_name="df_renamer",
        project_dir=str(datafiles),
        sheet_config_dir=str(datafiles),
        profile_dir=str(datafiles),
    )
    project = Project(flags, "sheetwork_test")
    config = ConfigLoader(flags, project)
    profile = Profile(project)
    with mock.patch.object(
        SheetBag, "_obtain_googlesheet", return_value=generate_test_df(DIRTY_DF)
    ):
        sheetbag = SheetBag(config, flags, profile)
        sheetbag.load_sheet()
        target_df = generate_test_df(RENAMED_DF)
        assert target_df.equals(sheetbag.sheet_df)



import logging
from pathlib import Path


class LogManager:
    def __init__(
        self,
        log_file_path: Path = Path(Path.cwd(), "sheetwork_logs"),
        log_to_console: bool = True,
    ):
        Path(log_file_path).mkdir(parents=True, exist_ok=True)
        log_filename = Path(log_file_path, "sheetwork_log.log")
        logger = logging.getLogger("Sheetwork Logger")
        logger.setLevel(logging.INFO)
        # Create handlers
        f_handler = logging.FileHandler(log_filename)
        f_handler.setLevel(logging.DEBUG)

        # Create formatters and add it to handlers
        f_format = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s"
        )
        f_handler.setFormatter(f_format)

        # Add handlers to the logger
        logger.addHandler(f_handler)

        # if we want to print the log to console we're going to add a streamhandler
        if log_to_console:
            c_handler = logging.StreamHandler()
            c_handler.setLevel(logging.INFO)
            c_format = logging.Formatter("%(message)s")
            c_handler.setFormatter(c_format)
            logger.addHandler(c_handler)

        self.logger = logger

    def set_debug(self):
        self.logger.setLevel(logging.DEBUG)
        for handler in self.logger.handlers:
            handler.setLevel(logging.DEBUG)


log_manager = LogManager()

GLOBAL_LOGGER = log_manager.logger



from .mockers import DIRTY_DF, CLEAN_DF, CASING_DF, SNAKE_CASED_COLS, generate_test_df


def test_cleanup():
    from core.cleaner import SheetCleaner

    dirty_df = generate_test_df(DIRTY_DF)
    clean_df = SheetCleaner(dirty_df).cleanup()
    expected_df = generate_test_df(CLEAN_DF)

    assert clean_df.equals(expected_df)


def test_snake_to_camel():
    from core.cleaner import SheetCleaner

    cased_df = generate_test_df(CASING_DF)
    recased_df = SheetCleaner(cased_df, True).cleanup()
    print(recased_df.columns.tolist())

    assert recased_df.columns.tolist() == SNAKE_CASED_COLS



class FlagParser:
    """Holds flags from args or sets up default ones that are mainly used to testing and delaying
    argument parsing from CLI so that pytest doesn't steal them and thinks they're for him and
    ultimately complain. It's one way to do it... There are probably other but I feel ok with it.
    """

    def __init__(
        self,
        parser: "parser.ArgumentParser",  # type: ignore
        test_sheet_name: str = str(),
        sheet_config_dir: str = str(),
        profile_dir: str = str(),
        project_dir: str = str(),
        project_name: str = str(),
    ):
        self.sheet_name = test_sheet_name
        self.create_table = False
        self.sheet_key = str()
        self.target_schema = str()
        self.target_table = str()
        self.log_level = str()
        self.interactive = False
        self.dry_run = False
        self.parser = parser
        self.sheet_config_dir = sheet_config_dir
        self.profile_dir = profile_dir
        self.project_dir = project_dir
        self.target = "test"
        self.project_name = project_name
        self.force_credentials = False

    def consume_cli_arguments(self):
        self.args = self.parser.parse_args()
        self.task = self.args.command

        # these only come into the flags if task is upload so we have to jump over.
        # there might be a more clever way to do this but we'll see when we add some other tasks.
        if self.task == "upload":
            self.sheet_name = self.args.sheet_name
            self.create_table = self.args.create_table
            self.sheet_key = self.args.sheet_key
            self.target_schema = self.args.schema
            self.target_table = self.args.table
            self.interactive = self.args.interactive
            self.dry_run = self.args.dry_run
            self.sheet_config_dir = self.args.sheet_config_dir
            self.target = self.args.target
        if self.task == "init":
            self.project_name = self.args.project_name
            self.force_credentials = self.args.force_credentials_folders

        self.log_level = self.args.log_level
        self.profile_dir = self.args.profile_dir
        self.project_dir = self.args.project_dir



from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine

from core.config.profile import Profile
from core.exceptions import CredentialsParsingError


class Credentials:
    """This should be a base class down the line but for now it's kinda adhoc until all works fine.
    """

    def __init__(self, profile: Profile):
        self.profile: dict = profile.profile_dict
        self.are_valid_credentials: bool = False
        self.db_type: str = str()
        self.credentials: dict = dict()
        self.validate_credentials()
        self.parse_credentials()

    def validate_credentials(self):
        # check that all necessary keys are in the profile (nullity will have been handled by
        # the yaml validator upsteam)
        db_type = self.profile.get("db_type")
        if db_type == "snowflake":
            must_have_keys = {
                "account",
                "user",
                "password",
                "role",
                "database",
                "warehouse",
                "schema",
            }
            keys_missing = must_have_keys.difference(self.profile.keys())
            if keys_missing:
                raise CredentialsParsingError(
                    f"The following keys: {keys_missing} must be in your profile."
                )
            self.are_valid_credentials = True
            self.db_type = db_type

    def parse_credentials(self):
        if self.profile.get("db_type") == "snowflake":
            must_have_keys = {
                "account",
                "user",
                "password",
                "role",
                "database",
                "warehouse",
                "schema",
            }
            for key in must_have_keys:
                self.credentials.update({key: self.profile.get(key)})


class Connection:
    """For now this is also pretty adhocy until all works well and we can genericise into proper
    adaptor design.
    """

    def __init__(self, credentials: Credentials):
        self.db_type = credentials.db_type
        self.credentials = credentials
        self.generate_engine()

    def generate_engine(self):
        self.engine = create_engine(
            URL(
                account=self.credentials.credentials.get("account"),
                user=self.credentials.credentials.get("user"),
                password=self.credentials.credentials.get("password"),
                role=self.credentials.credentials.get("role"),
                warehouse=self.credentials.credentials.get("warehouse"),
                database=self.credentials.credentials.get("database"),
                schema=self.credentials.credentials.get("schema"),
            )
        )



from typing import Dict

import colorama

COLOURS: Dict[str, str] = {
    "red": colorama.Fore.RED,
    "green": colorama.Fore.GREEN,
    "yellow": colorama.Fore.YELLOW,
    "reset_all": colorama.Style.RESET_ALL,
}



from pathlib import Path

from core.config.project import Project
from core.exceptions import InvalidProfileError, ProfileParserError
from core.logger import GLOBAL_LOGGER as logger
from core.yaml.yaml_helpers import open_yaml, validate_yaml
from core.yaml.yaml_schema import profiles_schema


class Profile:
    """Load, validate and set profile for sheetwork.

    Raises:
        FileNotFoundError: If no profiles.yml can be found in the expected location.
        InvalidProfileError: When a profile is not valid to the requirements

    Returns:
        Profile: Class holding all profiles variables necessary for clients and connections to be
        setup.
    """

    def __init__(self, project: Project, target_name: str = str()):
        self.profile_name = project.project_name
        self.target_name = target_name
        self.profile_dict: dict = dict()
        self.cannot_be_none = {"db_type", "guser"}
        self.profile_dir: Path = project.profile_dir
        self.google_credentials_dir = Path(project.profile_dir, "google").resolve()
        self.read_profile()
        logger.debug(f"PROFILE_DIR {self.profile_dir}")
        logger.debug(f"PROFILE_NAME: {self.profile_name}")

    def read_profile(self):
        logger.debug(f"Profile Name: {self.profile_name}")
        filename = Path(self.profile_dir, "profiles.yml")
        if filename.exists():
            yaml_dict = open_yaml(filename)
            is_valid_yaml = validate_yaml(yaml_dict, profiles_schema)
            profile = yaml_dict["profiles"].get(self.profile_name)
            if profile:
                # set target name from profile unless one was given at init from flags parse.
                if not self.target_name:
                    self.target_name = profile.get("target")
                if profile.get("outputs"):
                    target_profile = profile["outputs"].get(self.target_name)
                if target_profile and is_valid_yaml:
                    is_valid_profile = self._validate_profile(target_profile)
                    if is_valid_profile:
                        self.profile_dict = target_profile
                else:
                    raise ProfileParserError(
                        f"Error finding and entry for {self.target_name}, for {self.profile_name}."
                    )
            else:
                raise ProfileParserError(f"Could not find an entry for {self.profile_name}")
        else:
            raise FileNotFoundError(
                f"Could not open or find {filename.resolve()} check that it exists"
            )

    def _validate_profile(self, profile_dict: dict) -> bool:
        if isinstance(profile_dict, dict):
            keys_with_nones = {k for k, v in profile_dict.items() if not v}
            keys_with_nones.intersection_update(self.cannot_be_none)
            if keys_with_nones:
                raise InvalidProfileError(
                    f"The following fields: {keys_with_nones} cannot be empty."
                )
            return True
        raise ProfileParserError(f"Error finding and entry for {self.target_name}.")



config_schema = {
    "sheets": {
        "required": True,
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "sheet_name": {"required": True, "type": "string"},
                "sheet_key": {"required": True, "type": "string"},
                "worksheet": {"required": False, "type": "string"},
                "target_schema": {"required": False, "type": "string"},
                "target_table": {"required": True, "type": "string"},
                "snake_case_camel": {"required": False, "type": "boolean"},
                "columns": {
                    "type": "list",
                    "required": False,
                    "schema": {
                        "type": "dict",
                        "schema": {
                            "name": {
                                "required": True,
                                "type": "string",
                                "maxlength": 255,
                            },
                            "datatype": {
                                "required": True,
                                "type": "string",
                                "regex": "(?i)^(int|varchar|numeric|boolean|date|timestamp_ntz)$",
                            },
                            "identifier": {"required": False, "type": "string"},
                        },
                    },
                },
                "excluded_columns": {
                    "anyof_type": ["list", "string"],
                    "required": False,
                    "schema": {"type": "string"},
                },
            },
        },
    }
}

profiles_schema = {
    "profiles": {
        "required": True,
        "type": "dict",
        "valuesrules": {
            "type": "dict",
            "schema": {
                "target": {"required": True, "type": "string"},
                "outputs": {
                    "required": True,
                    "type": "dict",
                    "valuesrules": {
                        "type": "dict",
                        "schema": {
                            "db_type": {"required": True, "type": "string"},
                            "account": {"required": False, "type": "string"},
                            "user": {"required": True, "type": "string"},
                            "password": {"required": True, "type": "string"},
                            "role": {"required": False, "type": "string"},
                            "database": {"required": False, "type": "string"},
                            "warehouse": {"required": False, "type": "string"},
                            "schema": {"required": False, "type": "string"},
                            "guser": {"required": True, "type": "string"},
                        },
                    },
                },
            },
        },
    }
}

project_schema = {
    "name": {"required": True, "type": "string"},
    "target_schema": {"required": False, "type": "string"},
    "always_create": {"required": False, "type": "boolean"},
    "paths": {
        "type": "dict",
        "required": False,
        "schema": {
            "profile_dir": {"required": False, "type": "string"},
            "sheet_config_dir": {"required": False, "type": "string"},
        },
    },
}



from pathlib import Path

import yaml
from cerberus import Validator

from core.exceptions import SheetConfigParsingError, YAMLFileEmptyError


def open_yaml(path: "Path"):
    if path.is_file():
        with open(path, "r") as stream:
            yaml_file = yaml.safe_load(stream)  # type: ignore
            if yaml_file:
                return yaml_file
            raise YAMLFileEmptyError(f"Your yml file {path.resolve()} seems empty.")
    raise FileNotFoundError(f"File {path.resolve()} was not found.")


def validate_yaml(yaml: dict, validation_schema: dict) -> bool:
    v = Validator()
    valid_yaml = v.validate(yaml, validation_schema)
    if not valid_yaml:
        raise SheetConfigParsingError(f"scheet.yml is not formatted properly. \n {v.errors}")
    return valid_yaml



import sys
from typing import TYPE_CHECKING
from core.logger import GLOBAL_LOGGER as logger

if TYPE_CHECKING:
    from pathlib import Path


def open_dir_cmd() -> str:
    # https://docs.python.org/2/library/sys.html#sys.platform
    if sys.platform == "win32":
        return "start"

    if sys.platform == "darwin":
        return "open"

    return "xdg-open"


def make_dir(path: "Path"):
    logger.debug(f"Making folder: {path}")
    path.mkdir()


def make_file(path: "Path", contents: str = str()):
    logger.debug(f"Making file: {path}")
    path.touch()
    if contents:
        with path.open("w", encoding="utf-8") as f:
            f.write(contents)



from .mockers import CAST_DF, TO_CAST_DF, generate_test_df

CASTING_DICT = {
    "col_int": "int",
    "col_varchar": "varchar",
    "created_date": "date",
}


def test_cast_pandas_dtypes():
    from core.utils import cast_pandas_dtypes

    to_cast = generate_test_df(TO_CAST_DF)
    cast_df = cast_pandas_dtypes(to_cast, CASTING_DICT)
    expected_cast = generate_test_df(CAST_DF)

    assert cast_df.to_dict() == expected_cast.to_dict()


def test_check_columns_in_df():
    from core.utils import check_columns_in_df

    cols = ["column_that_is_not_in_df", "col_int"]

    to_check = generate_test_df(TO_CAST_DF)
    is_subset, columns = check_columns_in_df(to_check, cols, suppress_warning=True)

    assert is_subset is False
    assert columns == [cols[1]]


def test_check_dupe_cols():
    from core.utils import check_dupe_cols

    list_with_dupe = ["a", "a", "b"]
    dupes = check_dupe_cols(list_with_dupe, suppress_warning=True)

    assert dupes == ["a"]


def test_check_and_compare_version(mocker):
    from core.utils import check_and_compare_version

    # mock the call to pypi
    mocker.patch("luddite.get_version_pypi", return_value="1.0.0")
    dummy_version = "0.0.0"
    needs_update = check_and_compare_version(dummy_version)
    assert needs_update is True



import time
from pathlib import Path

from core.clients.system import make_dir, make_file, open_dir_cmd
from core.exceptions import MissnigInitProjectName, ProjectIsAlreadyCreated
from core.flags import FlagParser
from core.logger import GLOBAL_LOGGER as logger
from core.ui.printer import green

PROJECT_DOC_URL = "https://bastienboutonnet.gitbook.io/sheetwork/installation-and-configuration/untitled/set-up-your-sheetwork-project"
PROFILE_DOC_URL = "https://bastienboutonnet.gitbook.io/sheetwork/installation-and-configuration/untitled/set-up-your-sheetwork-profile"
GOOGLE_CREDS_DOC_URL = "https://bastienboutonnet.gitbook.io/sheetwork/installation-and-configuration/untitled/connecting-to-google-sheets"
SHEETS_CONFIG_DOC_URL = "https://bastienboutonnet.gitbook.io/sheetwork/usage/sheet-configuration"

INIT_DONE = """
Your new sheetwork project "{project_name}" has been created ✨.

Here is what happened behind the scenes:
- {project_path} was created.
- Inside that project, we created "sheetwork_project.yml" containing the bare essentials to get you started.
- An empty google credentials file was dropped in {google_path} and we called it {project_name}.json
- If it was your first time setting up sheetwork on your machine, we also created a profiles.yml file

What you need to do now:
{to_do_credentials}

Optionally:
- You might want to change some defaults in your {project_path}/{project_name}.yml file. For help:
    {project_doc_url}

- You might want to configure a sheet to import in your sheets.yml file. For help:
    {sheets_config_doc_url}
"""

CREDENTIALS_ONLY_SUCCESS_MESSAGE = """
Alright! Your credential and profile files have been creeated ✨

What you need to do now:
{to_do_credentials}
"""

TO_DO_CREDENTIALS = """
- Fill up your profiles.yml file. You can access it by running the following command:

    {open_cmd} {profiles_path}

- For help on how to fill your profiles.yml file head over to:
    {profiles_doc_url}

- You will need to fill up the {project_name}.json file with your google credentials key. For help see:
    {google_creds_doc_url}
"""

PROFILES_PATH = Path("~/.sheetwork/").expanduser()
PROJECT_PATH = Path.cwd()

PROJECT_FILE = """
name: '{project_name}'

# change the following to your default destination schema
target_schema: 'sandbox'

# we set sheetwork to always create tables, feel free to set that to false if you don't like it.
always_create: true
"""


class InitTask:
    def __init__(self, flags: FlagParser):
        self.flags = flags
        self.project_name: str = flags.project_name
        self.profiles_path: Path = PROFILES_PATH
        self.project_path: Path = PROJECT_PATH
        self.google_path: Path = Path()
        self.project_dir_is_created = False

    def assert_project_name(self):
        if not self.flags.project_name:
            raise MissnigInitProjectName(
                f"Please provide a project name to init your project with."
            )

    def override_paths(self):
        if self.flags.profile_dir:
            self.profiles_path = Path(self.flags.profile_dir)

        if self.flags.project_dir:
            self.project_path = Path(self.flags.project_dir)

    def create_profiles_dir(self):
        if not self.profiles_path.exists():
            make_dir(self.profiles_path)
        else:
            logger.debug(f"{self.profiles_path} already exists.")

    def create_profiles_file(self):
        profile_file = Path(self.profiles_path, "profiles").with_suffix(".yml")
        if not profile_file.exists():
            make_file(profile_file)
        else:
            logger.debug(f"{profile_file} already exists.")

    def create_google_dir_and_file(self):
        self.google_path = self.profiles_path / "google"
        google_file = self.google_path / f"{self.project_name}.json"

        if not self.google_path.exists():
            make_dir(self.google_path)
        else:
            logger.debug(f"{self.google_path} already exists.")

        if not google_file.exists():
            make_file(google_file)
        else:
            logger.debug(f"{google_file} already exists.")

    def create_project_dir(self):
        project_dir = self.project_path / f"{self.project_name}"
        if not project_dir.exists():
            make_dir(project_dir)
            self.project_dir_is_created = True
        elif self.flags.force_credentials:
            logger.warn(f"{self.project_name} already exists, moving on to credential files.")
        else:
            raise ProjectIsAlreadyCreated(
                f"""\n
                {self.project_name} already exists, so we'll stop.
                If you created it by mistake, delete it and run this again.

                If you want to generate the profiles and credentials files only use
                --force-credentials-folders CLI arguments (see help for more info).
                """
            )

    def create_project_file(self):
        full_path = Path(self.project_path, self.project_name, "sheetwork_project").with_suffix(
            ".yml"
        )
        if not full_path.exists():
            project_file_content = PROJECT_FILE.format(project_name=self.project_name)
            make_file(full_path, project_file_content)

    def show_complete(self):
        credentials_message = TO_DO_CREDENTIALS.format(
            to_do_credentials=TO_DO_CREDENTIALS,
            open_cmd=open_dir_cmd(),
            profiles_path=self.profiles_path,
            profiles_doc_url=PROFILE_DOC_URL,
            google_creds_doc_url=GOOGLE_CREDS_DOC_URL,
            project_name=self.project_name,
        )
        if self.project_dir_is_created:
            done_message = INIT_DONE.format(
                project_name=self.project_name,
                project_path=self.project_path,
                profiles_path=self.profiles_path,
                google_path=self.google_path,
                profiles_doc_url=PROFILE_DOC_URL,
                google_creds_doc_url=GOOGLE_CREDS_DOC_URL,
                project_doc_url=PROJECT_DOC_URL,
                sheets_config_doc_url=SHEETS_CONFIG_DOC_URL,
                to_do_credentials=credentials_message,
                open_cmd=open_dir_cmd(),
            )
        else:
            done_message = CREDENTIALS_ONLY_SUCCESS_MESSAGE.format(
                to_do_credentials=credentials_message
            )
        logger.info(green(done_message))

    def run(self):
        # print something cos it's fun!
        print(
            """
           ______           __                  __
          / __/ /  ___ ___ / /__    _____  ____/ /__
         _\ \/ _ \/ -_) -_) __/ |/|/ / _ \/ __/  '_/
        /___/_//_/\__/\__/\__/|__,__/\___/_/ /_/\_\\
        """
        )
        logger.info("Alright let's get to work")
        logger.info("❤️ Taking peanut butter and jelly out of the cupboard 🍇")
        time.sleep(3)

        # do the actual work people cared about in the first place.
        self.assert_project_name()
        self.override_paths()
        self.create_project_dir()
        self.create_project_file()
        self.create_profiles_dir()
        self.create_profiles_file()
        self.create_google_dir_and_file()
        self.show_complete()



from pathlib import Path
from typing import TYPE_CHECKING

from core.config.project import Project
from core.exceptions import (
    SheetConfigParsingError,
    SheetWorkConfigMissingError,
    TargetSchemaMissing,
)
from core.logger import GLOBAL_LOGGER as logger
from core.ui.printer import yellow
from core.yaml.yaml_helpers import open_yaml, validate_yaml
from core.yaml.yaml_schema import config_schema

if TYPE_CHECKING:
    from core.flags import FlagParser


class ConfigLoader:
    def __init__(self, flags: "FlagParser", project: "Project"):
        self.config: dict = dict()
        self.sheet_config: dict = dict(
            sheet_key=flags.sheet_key,
            target_schema=flags.target_schema,
            target_table=flags.target_table,
        )
        self.target_schema = flags.target_schema
        self.target_table = flags.target_table
        self.sheet_column_rename_dict: dict = dict()
        self.sheet_columns: dict = dict()
        self.excluded_columns: list = list()
        self.flags = flags
        self.project = project
        self.yml_folder: Path = project.sheet_config_dir
        logger.debug(f"SHEET_FOLDER: {project.sheet_config_dir}")
        self.set_config()

    def set_config(self):
        if self.flags.sheet_name:
            self.load_config_from_file()
        elif self.flags.sheet_key and self.flags.target_schema and self.flags.target_table:
            logger.debug("Reading config from command line.")
        else:
            raise NotImplementedError(
                """
                No target schema and or target was provided.
                You must provide one when not reading from config file.
                """
            )

    def load_config_from_file(self):
        logger.debug("Reading config from config file.")
        filename = Path(self.yml_folder, "sheets.yml")
        logger.debug(f"SHEET FILENAME: {filename}")
        if filename.exists():
            config_yaml = open_yaml(filename)
        else:
            raise SheetWorkConfigMissingError(
                """
                Are you in a sheetwork folder? Cannot find 'sheets.yml' to import config from.
                If you plan to run sheetwork from a different folder than current you'll have to
                provide a custom path to the config files. See --help for arguments."""
            )
        if config_yaml:
            is_valid_yaml = validate_yaml(config_yaml, config_schema)
        if is_valid_yaml:
            self.config = config_yaml
            self.get_sheet_config()
            self._generate_column_type_override_dict()
            self._generate_column_rename_dict()
            self._override_cli_args()

    @staticmethod
    def lowercase(obj):
        """ Make dictionary lowercase """
        new = dict()
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, dict):
                    v = lowercase(v)
                if k == "name":
                    new[k] = v.lower()
                else:
                    new[k] = v
        return new

    def get_sheet_config(self):
        if self.flags.sheet_name:
            sheets = self.config["sheets"]
            sheet_config = [
                sheet for sheet in sheets if sheet["sheet_name"] == self.flags.sheet_name
            ]
            if len(sheet_config) > 1:
                raise SheetConfigParsingError(
                    f"Found more than one config for {self.flags.sheet_name}. "
                    "Check your sheets.yml file."
                )
            if not sheet_config:
                raise SheetConfigParsingError(
                    f"No configuration was found for {self.flags.sheet_name}. "
                    "Check your sheets.yml file."
                )
            self.sheet_config = sheet_config[0]
            logger.debug(f"Sheet config dict: {self.sheet_config}")
            if self.sheet_config.get("columns"):
                self.sheet_config["columns"] = [
                    self.lowercase(column_dict) for column_dict in self.sheet_config.get("columns")
                ]
                logger.debug(f"Cols after lowercasing: {self.sheet_config.get('columns')}")
        else:
            raise SheetWorkConfigMissingError("No sheet name was provided, cannot fetch config.")

    def _generate_column_type_override_dict(self):
        """Generates a dictionary of key, value where key is the name of a column and value is the
        name of the datatype into that column should be cast on table creation.
        """

        try:
            if self.sheet_config and self.sheet_config.get("columns"):
                columns = self.sheet_config.get("columns")
                column_dict = dict()
                for column in columns:
                    column_dict.update(dict({column.get("name"): column.get("datatype")}))
                if column_dict:
                    logger.debug(f"colums operations dict: {column_dict}")
                    self.sheet_columns = column_dict
        except KeyError as e:
            logger.warning(
                yellow(
                    f"No {str(e)} data for {self.flags.sheet_name}. But that might be intentional."
                )
            )

    def _generate_column_rename_dict(self):
        """Generates a dictionary of key values where key is the original name in the sheet and
        value is the target name.
        """

        if self.sheet_config:
            columns = self.sheet_config.get("columns")
            if columns:
                column_rename_dict = dict()
                for column in columns:
                    if column.get("identifier"):
                        column_rename_dict.update(dict({column["identifier"]: column["name"]}))
                    if column_rename_dict:
                        logger.debug(f"column renaming dict {column_rename_dict}")
                        self.sheet_column_rename_dict = column_rename_dict

    def _override_cli_args(self):
        """Overrides any CLI argument that may have been passed to sheeload when it reads from a
        config yaml file, thereby giving precedence to the arguments in the .yml file.
        """

        self.sheet_key = self.sheet_config["sheet_key"]
        if not self.target_table:
            self.target_table = self.sheet_config.get("target_table")
        if not self.target_schema:
            self.target_schema = self.sheet_config.get("target_schema", self.project.target_schema)
        if not self.target_schema:
            raise TargetSchemaMissing(
                "No target schema found. You must provide one either in the project, sheet or CLI"
            )



import os

import pytest

from core.flags import FlagParser

from .mockers import EXPECTED_CONFIG, NO_COLS_EXPECTED_CONFIG

FIXTURE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)))


@pytest.mark.datafiles(FIXTURE_DIR)
def test_set_config(datafiles):
    from core.config.config import ConfigLoader

    from core.config.project import Project
    from core.main import parser

    flags = FlagParser(
        parser,
        test_sheet_name="df_dropper",
        project_dir=str(datafiles),
        sheet_config_dir=str(datafiles),
    )
    project = Project(flags)
    config = ConfigLoader(flags, project)

    flags2 = FlagParser(
        parser,
        test_sheet_name="no_cols",
        project_dir=str(datafiles),
        sheet_config_dir=str(datafiles),
    )
    project2 = Project(flags2)
    config2 = ConfigLoader(flags2, project2)

    assert config.sheet_config == EXPECTED_CONFIG
    assert config2.sheet_config == NO_COLS_EXPECTED_CONFIG


@pytest.mark.datafiles(FIXTURE_DIR)
def test__override_cli_args(datafiles):
    from core.config.config import ConfigLoader

    from core.config.project import Project
    from core.main import parser

    flags = FlagParser(
        parser,
        test_sheet_name="sheet_with_no_schema",
        project_dir=str(datafiles),
        sheet_config_dir=str(datafiles),
    )
    project = Project(flags)
    config = ConfigLoader(flags, project)
    assert config.target_schema == project.target_schema



import collections
from pathlib import Path
from typing import Optional, Tuple, Union
from urllib.error import URLError

import luddite
import pandas
from packaging.version import parse as semver_parse

from core._version import __version__
from core.exceptions import (
    ColumnNotFoundInDataFrame,
    DuplicatedColumnsInSheet,
    NearestFileNotFound,
    UnsupportedDataTypeError,
)
from core.logger import GLOBAL_LOGGER as logger
from core.ui.printer import yellow


class PathFinder:
    """Finds paths of files by going up a number of parent folders. That's it really. It's a class
    mainly to be able to limit the number of iterations otherwise we could end up going up and
    up and up and up...

    Raises:
        NearestFileNotFound: When the required file in find_nearest_dir_and_file() cannot be found
        within the maximum number of allowed iterations upwards.

    Returns:
        PathFinder: Yeah nothing much more, real returns come from the class methods.
    """

    def __init__(self, max_iter: int = 4):
        self.max_iter = max_iter
        self.iteration = 0

    def find_nearest_dir_and_file(
        self, yaml_file: str, current: Path = Path.cwd()
    ) -> Tuple[Path, Path]:
        """Looks for the yaml_file you ask for starting from the current directory and going up with
        recursion while the iteration number is still within the max allowed.

        Args:
            yaml_file (str): Name and extension of the file to find.
            current (Path, optional): Path() objects from which to start. Defaults to Path.cwd().

        Raises:
            NearestFileNotFound: When no file that matches the required name can be found.

        Returns:
            Tuple[Path, Path]: The directory up to the file name, and the full path to the filename,
            respectively. Maybe we'll end up deprecating one of these returns down the line but for
            now it's handy.
        """
        filename = Path(current, yaml_file)
        while self.iteration < self.max_iter:
            logger.debug(f"Looking for {filename}")
            if filename.exists():
                project_dir = filename.parent
                logger.debug(f"{filename} exists and was returned")
                return project_dir, filename
            current = current.parent
            filename = Path(current, yaml_file)
            self.iteration += 1
        else:
            raise NearestFileNotFound(
                f"Unable to find {yaml_file} in the nearby directories after {self.max_iter} "
                "iterations upwards."
            )


def cast_pandas_dtypes(df: pandas.DataFrame, overwrite_dict: dict = dict()) -> pandas.DataFrame:
    overwrite_dict = overwrite_dict.copy()
    dtypes_map = dict(
        varchar="object",
        int="int64",
        numeric="float64",
        boolean="bool",
        timestamp_ntz="datetime64[ns]",
        date="datetime64[ns]",  # this intentional pandas doesn't really have just dates.
    )

    # Check for type support
    unsupported_dtypes = set(overwrite_dict.values()).difference(dtypes_map.keys())
    if unsupported_dtypes:
        raise UnsupportedDataTypeError(f"{unsupported_dtypes} are currently not supported")

    # check overwrite col is in df
    invalid_columns = set(overwrite_dict.keys()).difference(set(df.columns.tolist()))
    if invalid_columns:
        raise ColumnNotFoundInDataFrame(f"{invalid_columns} not in DataFrame. Check spelling?")

    # recode dict in pandas terms
    for col, data_type in overwrite_dict.items():
        overwrite_dict.update({col: dtypes_map[data_type]})

    # cast
    df = df.astype(overwrite_dict)
    logger.debug(f"Head of cast dataframe:\n {df.head()}")
    return df


def check_columns_in_df(
    df: pandas.DataFrame,
    columns: Union[list, str],
    warn_only: bool = False,
    suppress_warning: bool = False,
) -> Tuple[bool, list]:
    if isinstance(columns, str):
        columns = [columns]
    is_subset = set(columns).issubset(df.columns)
    if is_subset:
        return True, columns
    # else reduce columms, provide filtered list set bool to false and warn or raise
    cols_not_in_df = [x for x in columns if x not in df.columns.tolist()]
    reduced_cols = [x for x in columns if x in df.columns.tolist()]
    message = f"The following columns were not found in the sheet: {cols_not_in_df} "
    if warn_only and not suppress_warning:
        logger.warning(
            yellow(message + "they were ignored. Consider cleaning your sheets.yml file")
        )
    elif not warn_only and not suppress_warning:
        raise ColumnNotFoundInDataFrame(message + "Google Sheet or sheets.yml needs to be cleaned")
    return False, reduced_cols


def check_dupe_cols(columns: list, suppress_warning: bool = False) -> Optional[list]:
    """checks dupes in a list"""
    dupes = [item for item, count in collections.Counter(columns).items() if count > 1]
    if dupes and not suppress_warning:
        raise DuplicatedColumnsInSheet(
            f"Duplicate column names found in Google Sheet: {dupes}. Aborting. Fix your sheet."
        )
    return dupes


def check_and_compare_version(external_version: Optional[str] = str()) -> bool:
    try:
        pypi_version = luddite.get_version_pypi("sheetwork")
        if external_version:
            installed_version = external_version
        else:
            installed_version = __version__

        needs_update = semver_parse(pypi_version) > semver_parse(installed_version)
        if needs_update:
            logger.warn(
                yellow(
                    f"Looks like you're a bit behind. A newer version of Sheetwork v{pypi_version} is available."
                )
            )
        return needs_update

    except URLError:
        return False



from pathlib import Path

import pytest

from .mockers import EXPECTED_DEV_TEST_PROFILE

FIXTURE_DIR = Path(__file__).resolve().parent


@pytest.mark.datafiles(FIXTURE_DIR)
def test_read_profile(datafiles):
    from core.config.profile import Profile
    from core.config.project import Project
    from core.flags import FlagParser
    from core.main import parser

    print(FIXTURE_DIR)
    flags = FlagParser(parser, project_dir=str(datafiles), profile_dir=str(datafiles))
    project = Project(flags, "sheetwork_test")
    profile = Profile(project, "dev")
    profile.read_profile()

    assert profile.profile_dict == EXPECTED_DEV_TEST_PROFILE



from pathlib import Path

import gspread
import pandas
from gspread.exceptions import SpreadsheetNotFound, WorksheetNotFound
from oauth2client.service_account import ServiceAccountCredentials

from core.config.profile import Profile
from core.exceptions import (
    GoogleCredentialsFileMissingError,
    GoogleSpreadSheetNotFound,
    NoWorkbookLoadedError,
    WorksheetNotFoundError,
)
from core.logger import GLOBAL_LOGGER as logger
from core.ui.printer import green
from core.utils import check_dupe_cols

SCOPE = [
    "https://spreadsheets.google.com/feeds",
    "https://www.googleapis.com/auth/drive",
]


class GoogleSpreadsheet:
    """Gets google cloud credentials checks client from profile and handles google sheet
    interactions such as downloading a sheet or other activities permitted by the gspread lib.

    Raises:
        GoogleCredentialsFileMissingError: When the credentials file cannot be located in the expect
        location.
        GoogleSpreadSheetNotFound: When the required google sheet cannot be found.

    Returns:
        GoogleSpreadsheet: GoogleSpreadsheet object and methods.
    """

    def __init__(self, profile: Profile, workbook_key: str = str(), workbook_name: str = str()):
        p = Path(profile.google_credentials_dir, profile.profile_name).with_suffix(".json")
        if p.exists():
            self.credentials = ServiceAccountCredentials.from_json_keyfile_name(p, SCOPE)
            self.gc = gspread.authorize(self.credentials)
        else:
            raise GoogleCredentialsFileMissingError(
                "Sheetwork could not find a credentials file for your "
                f"'{profile.profile_name}' profile in the ~/.sheetwork/google/ folder. "
                "Check installation instructions if you do not know how to set this up."
            )
        self.workbook_key = workbook_key
        self.workbook_name = workbook_name
        self.client = profile.profile_dict.get("guser")
        self._open_workbook()

    def _open_workbook(self):
        try:
            if self.workbook_key:
                self.workbook = self.gc.open_by_key(self.workbook_key)
            elif self.workbook_name:
                self.workbook = self.gc.open(self.workbook_name)

        except SpreadsheetNotFound:
            raise GoogleSpreadSheetNotFound(
                "Spreadsheet not found. You either have a typo in the key or name provided "
                f"or your client {self.client} does not have read access to the sheet."
            )

    def make_df_from_worksheet(
        self, worksheet_name: str = str(), grab_header: bool = True
    ) -> pandas.DataFrame:
        if not self.workbook:
            raise NoWorkbookLoadedError(
                "Workbook object seems empty, cannot turn a None object into a dataframe"
            )
        try:
            if worksheet_name:
                worksheet = self.workbook.worksheet(worksheet_name)
            else:
                worksheet_name = "default sheet"
                worksheet = self.workbook.get_worksheet(0)
            logger.debug(green("Sheet loaded successfully"))
            if grab_header:
                values = worksheet.get_all_values()
                check_dupe_cols(values[0])
                df = pandas.DataFrame(values[1:], columns=values[0])
            else:
                df = pandas.DataFrame(worksheet.get_all_values())
            return df
        except WorksheetNotFound:
            raise WorksheetNotFoundError(
                f"Could not find {worksheet_name} in workbook. "
                "If 'default sheet' not found all sheets in the workbook may be empty."
            )



from colorama import Fore


class SheetWorkError(Exception):
    def __init__(self, message):
        super().__init__(Fore.RED + message)


class SheetWorkConfigMissingError(SheetWorkError):
    "When a sheet config cannot be found"


class GoogleCredentialsFileMissingError(SheetWorkError):
    "When the google credentials file could not be located."


class GoogleSpreadSheetNotFound(SheetWorkError):
    "When a referred spreadsheet cannot be found"


class NoWorkbookLoadedError(SheetWorkError):
    "When the workbook object is None"


class WorksheetNotFoundError(SheetWorkError):
    "when a referred worksheet cannot be found in the workbook"


class YAMLFileEmptyError(SheetWorkError):
    "When a yaml that exists returns nothing"


class SheetConfigParsingError(SheetWorkError):
    "For cases where sheet was found but content could not be parsed."


class ProfileParserError(SheetWorkError):
    "When no dict or an invalid dict came out of the profile reader"


class ProjectFileParserError(SheetWorkError):
    "When no dict comes out of loading project or other less specific project parsing related stuff"


class ColumnNotFoundInDataFrame(SheetWorkError):
    "For cases where renamer is provided the wrong identifier name."


class InvalidProfileError(SheetWorkError):
    "For when some values in a profile that should not be none are."


class NearestFileNotFound(SheetWorkError):
    "When the path finder has reached its max iterations without finding the expected file."


class CredentialsParsingError(SheetWorkError):
    "When the credentials parser cannot find the right keys and other nasties."


class UnsupportedDataTypeError(SheetWorkError):
    "When a requested cast isn't supported by the database or mapping enforced by SheetWork."


class DatabaseError(SheetWorkError):
    "To catch db interaction errors"


class TableDoesNotExist(SheetWorkError):
    "When query for rows and cols came back empty or none"


class DuplicatedColumnsInSheet(SheetWorkError):
    "when a google sheet contains the same column name twice"


class TargetSchemaMissing(SheetWorkError):
    "when no target schema whatsoever can be found"


class ProjectIsAlreadyCreated(SheetWorkError):
    "when using sheetwork init and a project is found"


class MissnigInitProjectName(SheetWorkError):
    "when no project name is requested"



import time
from core.ui.colours import COLOURS


USE_COLOURS = True
PRINTER_WIDTH = 80

CL_RED = COLOURS["red"]
CL_GREEN = COLOURS["green"]
CL_YELLOW = COLOURS["yellow"]
CL_RESET_ALL = COLOURS["reset_all"]


def colour(message: str, colour: str) -> str:
    if USE_COLOURS:
        return f"{colour}{message}{CL_RESET_ALL}"
    return message


def green(message: str):
    return colour(message, CL_GREEN)


def yellow(message: str):
    return colour(message, CL_YELLOW)


def red(message: str):
    return colour(message, CL_RED)


def timed_message(message: str) -> str:
    t = time.localtime()
    now = time.strftime("%H:%M:%S", t)
    return f"{now} ---- {message}"



import pandas
from pandas import Timestamp

EXPECTED_CONFIG = {
    "sheet_name": "df_dropper",
    "sheet_key": "sample",
    "target_schema": "sand",
    "target_table": "bb_test_sheetwork",
    "columns": [
        {"name": "col_a", "datatype": "int"},
        {"name": "col_b", "datatype": "varchar"},
        {"name": "col_one", "datatype": "varchar"},
        {"name": "renamed_col", "identifier": "long ass name", "datatype": "varchar"},
    ],
    "excluded_columns": ["to_exclude"],
}

EXPECTED_DEV_TEST_PROFILE = {
    "db_type": "snowflake",
    "account": "a",
    "user": "b",
    "password": "c",
    "role": "d",
    "database": "e",
    "warehouse": "f",
    "schema": "g",
    "guser": "sheetwork_test@blahh.iam.gserviceaccount.com",
}

NO_COLS_EXPECTED_CONFIG = {
    "sheet_name": "no_cols",
    "sheet_key": "sample",
    "target_schema": "sand",
    "target_table": "bb_test_sheetwork",
}

EXPECTED_SHEETWORK_PROJECT = {
    "name": "sheetwork_test",
    "target_schema": "sand",
    "always_create": True,
}

DIRTY_DF = {
    "col_a": [1, 2, 32],
    "col b": ["as .    ", "b", "   c"],
    "1. col_one": ["aa", "bb", "cc"],
    "": ["q", "q", "q"],
    "col_1": [1, 2, 33],
    "long ass name": ["foo", "bar", "fizz"],
}

TO_CAST_DF = {
    "col_int": ["1", "2", "32"],
    "col_varchar": ["foo", "bar", "fizz"],
    "created_date": ["2019/01/01", "2019/01/02", "2019/01/03"],
}

CAST_DF = {
    "col_int": {0: 1, 1: 2, 2: 32},
    "col_varchar": {0: "foo", 1: "bar", 2: "fizz"},
    "created_date": {
        0: Timestamp("2019-01-01 00:00:00"),
        1: Timestamp("2019-01-02 00:00:00"),
        2: Timestamp("2019-01-03 00:00:00"),
    },
}

CASING_DF = {
    "CamelCasedCol": [1, 2, 3],
    "snake_cased_col": [1, 2, 3],
}

SNAKE_CASED_COLS = ["camel_cased_col", "snake_cased_col"]

CAMEL_CASED_COLS = ["CamelCasedCol", "SnakeCasedCol"]

CLEAN_DF = {
    "col_a": {0: 1, 1: 2, 2: 32},
    "col_b": {0: "as .", 1: "b", 2: "c"},
    "col_one": {0: "aa", 1: "bb", 2: "cc"},
    "col_1": {0: 1, 1: 2, 2: 33},
    "long_ass_name": {0: "foo", 1: "bar", 2: "fizz"},
}

RENAMED_DF = {
    "col_a": {0: 1, 1: 2, 2: 32},
    "col_b": {0: "as .", 1: "b", 2: "c"},
    "col_one": {0: "aa", 1: "bb", 2: "cc"},
    "col_1": {0: 1, 1: 2, 2: 33},
    "renamed_col": {0: "foo", 1: "bar", 2: "fizz"},
}

DROP_COL_DF = {
    "col_a": [1, 2, 32],
    "col b": ["as .    ", "b", "   c"],
    "1. col_one": ["aa", "bb", "cc"],
    "": ["q", "q", "q"],
    "long ass name": ["foo", "bar", "fizz"],
    "to_exclude": ["garbage1", "garbage2", "garbage3"],
}

RENAMED_COLS = ["col_a", "col b", "1. col_one", "", "col_1", "renamed_col"]

EXCLUDED_DF_COLS = ["col_a", "col b", "1. col_one", "", "long ass name"]


def generate_test_df(df):
    test_df = pandas.DataFrame.from_dict(df)
    return test_df



from pathlib import Path
from typing import TYPE_CHECKING

from core.exceptions import ProjectFileParserError
from core.logger import GLOBAL_LOGGER as logger
from core.utils import PathFinder
from core.yaml.yaml_helpers import open_yaml, validate_yaml
from core.yaml.yaml_schema import project_schema

if TYPE_CHECKING:
    from core.flags import FlagParser

PROJECT_FILENAME = "sheetwork_project.yml"


class Project:
    """Sets up everything there is to know about the project config."""

    def __init__(self, flags: "FlagParser", project_name: str = str()):
        self.project_name = project_name
        self.project_dict: dict = dict()
        self.target_schema: str = str()
        self.always_create: bool = True
        self.flags = flags

        # directories (first overwritten by flags, then by project) This may not always be able to
        # be like this we might wanna give prio to CLI but for now this removes some complication.
        self.project_file_fullpath: Path = Path("dumpy_path")
        self.profile_dir: Path = Path("~/.sheetwork/").expanduser()
        self.sheet_config_dir: Path = Path.cwd()

        # override defaults
        self.override_from_flags()
        self.load_project_from_yaml()
        logger.debug(f"Project name: {self.project_name}")

    def load_project_from_yaml(self):
        if self.project_file_fullpath == Path("dumpy_path"):
            _, self.project_file_fullpath = PathFinder().find_nearest_dir_and_file(PROJECT_FILENAME)
        project_yaml = open_yaml(self.project_file_fullpath)
        is_valid_yaml = validate_yaml(project_yaml, project_schema)
        if project_yaml and is_valid_yaml:
            self.project_dict = project_yaml
            self.project_name = project_yaml.get("name", self.project_name)
            self.target_schema = project_yaml.get("target_schema", self.target_schema)
            if project_yaml.get("paths"):
                self.profile_dir = (
                    Path(project_yaml["paths"].get("profile_dir", self.profile_dir))
                    .expanduser()
                    .resolve()
                )
                self.sheet_config_dir = (
                    Path(project_yaml["paths"].get("sheet_config_dir", self.sheet_config_dir))
                    .expanduser()
                    .resolve()
                )
            self.always_create = project_yaml.get("always_create", self.always_create)
        else:
            raise ProjectFileParserError(
                f"Error trying to load project config from {self.project_file_fullpath}. "
                "Check it exists or that it is valid."
            )

    def override_from_flags(self):
        if self.flags.project_dir:
            self.project_file_fullpath = Path(self.flags.project_dir, PROJECT_FILENAME)
        if self.flags.profile_dir:
            self.profile_dir = Path(self.flags.profile_dir)
        if self.flags.sheet_config_dir:
            self.sheet_config_dir = Path(self.flags.sheet_config_dir)



import sys
from typing import TYPE_CHECKING, Tuple

import pandas

from core.adapters.connection import Connection, Credentials
from core.adapters.impl import SnowflakeAdapter
from core.cleaner import SheetCleaner
from core.clients.google import GoogleSpreadsheet
from core.config.config import ConfigLoader
from core.config.profile import Profile
from core.logger import GLOBAL_LOGGER as logger
from core.ui.printer import yellow, red, timed_message
from core.utils import check_columns_in_df

if TYPE_CHECKING:
    from core.flags import FlagParser


class SheetBag:
    """Main object orchestrates sheet loading, cleaning, and db pushing by calling the relevant
    submodules.

    Raises:
        TypeError: [description]
        ColumnNotFoundInDataFrame: If a column on which a rename or casting is asked for cannot be
        found in the DataFrame resulting from the obtained sheet.
        RuntimeError: [description]

    Returns:
        SheetBag: Loaded, and possibly cleaned sheet object with db interaction methods.
    """

    def __init__(self, config: "ConfigLoader", flags: "FlagParser", profile: "Profile"):
        self.sheet_df: pandas.DataFrame = pandas.DataFrame()
        self.flags = flags
        self.config = config
        self.target_schema = config.target_schema
        self.target_table = config.target_table
        self.profile = profile
        self.push_anyway = False

    def _obtain_googlesheet(self):
        worksheet = self.config.sheet_config.get("worksheet", str())
        df = GoogleSpreadsheet(
            self.profile, self.config.sheet_config["sheet_key"]
        ).make_df_from_worksheet(worksheet_name=worksheet)
        return df

    def load_sheet(self):
        """Loads a google sheet, and calls clean up steps if applicable.
        Sheet must have been shared with account admin email address used in storage.

        Raises:
            TypeError: When loader does not return results that can be converted into a pandas
            DataFrame a type error will be raised.
        """

        if self.flags.sheet_name:
            logger.info(timed_message(f"Importing: {self.flags.sheet_name}"))
            logger.debug(f"Importing data from: {self.config.sheet_config['sheet_key']}")
        else:
            logger.info(
                timed_message(f"Importing data from: {self.config.sheet_config.get('sheet_key')}")
            )
        df = self._obtain_googlesheet()
        if not isinstance(df, pandas.DataFrame):
            raise TypeError("import_sheet did not return a pandas DataFrame")
        logger.debug(f"Columns imported from sheet: {df.columns.tolist()}")

        # Perform exclusions, renamings and cleanups before releasing the sheet.
        df = self.exclude_columns(df)
        df = self.rename_columns(df)
        self.push_anyway, df = self.run_cleanup(df)
        logger.debug(f"Columns after cleanups and exclusions: {df.columns}")
        self.sheet_df = df

    def rename_columns(self, df):
        if self.config.sheet_column_rename_dict:
            _, _ = check_columns_in_df(df, self.config.sheet_column_rename_dict.keys())
            df = df.rename(columns=self.config.sheet_column_rename_dict)
        return df

    def exclude_columns(self, df: pandas.DataFrame) -> pandas.DataFrame:
        """Drops columns referred to by their identifier (the exact string in the google sheet) when
        a list is provided in the "excluded_columns" field of a sheet yml file.

        Args:
            df (pandas.DataFrame): DataFrame downloaded from google sheet.

        Returns:
            pandas.DataFrame: Either the same dataframe as originally provided or one with dropped
            columns as required.
        """

        cols_to_exclude = self.config.sheet_config.get("excluded_columns", list())
        if cols_to_exclude:
            _, filtered_columns = check_columns_in_df(df, cols_to_exclude, warn_only=True)
            if filtered_columns:
                df = df.drop(filtered_columns, axis=1)
            return df
        return df

    @staticmethod
    def _collect_and_check_answer(post_cleanup: bool = False):
        acceptable_answers = ["y", "n", "a"]
        user_input = str()
        while user_input not in acceptable_answers:
            if user_input is not None:
                logger.info("Choose 'y':yes, 'n':no, 'a':abort'")
            if post_cleanup:
                user_input = input("Would you like to push to db? (y/n):")
            else:
                user_input = input("Would you like to perform cleanup? (y/n/a): ")
        if user_input.lower() == "y":
            return True
        if user_input.lower() == "n":
            return False
        if user_input.lower() == "a":
            logger.info(red("User aborted."))
            sys.exit(1)

    @staticmethod
    def _show_dry_run_preview(sheet_df):

        print("\nDataFrame DataTypes: \n\n" + str(sheet_df.dtypes))
        print("\nDataFrame Header: \n\n" + str(sheet_df.head(10)))

    def run_cleanup(self, df) -> Tuple[bool, pandas.DataFrame]:
        clean_up = True
        # check for interactive mode
        if self.flags.interactive:
            logger.info(
                yellow(
                    "PRE-CLEANING PREVIEW: The DataFrame you would push to the database would look like this:"
                )
            )
            self._show_dry_run_preview(df)
            clean_up = self._collect_and_check_answer()

        if clean_up is True:
            logger.debug("Performing clean ups")
            clean_df = SheetCleaner(
                df, self.config.sheet_config.get("snake_case_camel", False)
            ).cleanup()
            if self.flags.dry_run or self.flags.interactive:
                logger.info(yellow("\nPOST-CLEANING PREVIEW:"))
                self._show_dry_run_preview(clean_df)
                carry_on = self._collect_and_check_answer(post_cleanup=True)
                if not carry_on:
                    logger.info(timed_message(red("User Aborted.")))
                    sys.exit(1)
            return True, clean_df
        return True, df

    def push_sheet(self):
        logger.info(timed_message("Pushing sheet to database..."))
        logger.debug(f"Column override dict is a {type(self.config.sheet_columns)}")
        logger.debug(f"Sheet columns: {self.config.sheet_columns}")
        logger.debug(f"Columns in final df: {self.sheet_df.columns.tolist()}")
        credentials = Credentials(self.profile)
        connection = Connection(credentials)
        adapter = SnowflakeAdapter(connection, self.config)
        adapter.upload(self.sheet_df, self.target_schema)

    def check_table(self):
        credentials = Credentials(self.profile)
        connection = Connection(credentials)
        adapter = SnowflakeAdapter(connection, self.config)
        adapter.check_table(self.target_schema, self.target_table)

    def run(self):
        self.load_sheet()
        if self.push_anyway:
            self.push_sheet()
            self.check_table()
        else:
            logger.info(yellow("Nothing pushed since you were in --dry_run mode."))



import tempfile
from typing import TYPE_CHECKING

import pandas
from sqlalchemy.types import BOOLEAN, DATE, INTEGER, TIMESTAMP, VARCHAR, Numeric

from core.config.config import ConfigLoader
from core.exceptions import DatabaseError, TableDoesNotExist
from core.logger import GLOBAL_LOGGER as logger
from core.ui.printer import green, timed_message
from core.utils import cast_pandas_dtypes

if TYPE_CHECKING:
    from core.adapters.connection import Connection


class SnowflakeAdapter:
    """Interacts with snowflake via SQLAlchemy"""

    def __init__(self, connection: "Connection", config: ConfigLoader):
        self.connection = connection
        self.engine = connection.engine
        self.config = config

    def acquire_connection(self):
        self.con = self.engine.connect()

    def close_connection(self):
        self.con.close()

    @staticmethod
    def sqlalchemy_dtypes(dtypes_dict) -> dict:
        dtypes_dict = dtypes_dict.copy()
        dtypes_map = dict(
            varchar=VARCHAR,
            int=INTEGER,
            numeric=Numeric(38, 18),
            boolean=BOOLEAN,
            timestamp_ntz=TIMESTAMP,
            date=DATE,
        )

        for col, data_type in dtypes_dict.items():
            dtypes_dict.update({col: dtypes_map[data_type]})
        return dtypes_dict

    def upload(self, df: pandas.DataFrame, override_schema: str = str()):
        # cast columns
        df = cast_pandas_dtypes(df, overwrite_dict=self.config.sheet_columns)
        dtypes_dict = self.sqlalchemy_dtypes(self.config.sheet_columns)

        # potenfially override target schema from config.
        if override_schema:
            schema = override_schema
        else:
            schema = self.config.target_schema

        # write to csv and try to talk to db
        temp = tempfile.NamedTemporaryFile()
        df.to_csv(temp.name, index=False, header=False, sep="|")

        try:
            self.acquire_connection()
            df.head(0).to_sql(
                name=self.config.target_table,
                schema=schema,
                con=self.con,
                if_exists="replace",
                index=False,
                dtype=dtypes_dict,
            )

            qualified_table = f"{self.config.target_schema}.{self.config.target_table}"
            self.con.execute(
                f"""
                create or replace temporary stage {self.config.target_table}_stg
                file_format = (type = 'CSV' field_delimiter = '|'
                skip_header = 0 field_optionally_enclosed_by = '"')
                """
            )
            self.con.execute(f"put file://{temp.name} @{self.config.target_table}_stg")
            self.con.execute(f"copy into {qualified_table} from @{self.config.target_table}_stg")
            self.con.execute(f"drop stage {self.config.target_table}_stg")
        except Exception as e:
            raise DatabaseError(e)
        finally:
            temp.close()
            self.close_connection()

    def execute(self, query: str, return_results: bool = False):
        self.acquire_connection()
        results = self.con.execute(query)
        if return_results:
            result_set = results.fetchall()
            self.close_connection()
            return result_set
        self.close_connection()
        return None

    def check_table(self, target_schema: str, target_table: str):
        columns_query = f"""
                select count(*)
                from dwh.information_schema.columns
                where table_catalog = 'DWH'
                and table_schema = '{target_schema.upper()}'
                and table_name = '{target_table.upper()}'
                ;
                """
        rows_query = rows_query = f"select count(*) from {target_schema}.{target_table}"
        columns = self.execute(columns_query, return_results=True)
        rows = self.execute(rows_query, return_results=True)
        if columns and rows:
            logger.info(
                timed_message(
                    green(
                        f"Push successful for "
                        f"{target_schema}.{target_table} \n"
                        f"Found {columns[0][0]} columns and {rows[0][0]} rows."
                    )
                )
            )
        else:
            raise TableDoesNotExist(f"Table {target_schema}.{target_table} seems empty")



from pathlib import Path

import pytest

from .mockers import EXPECTED_SHEETWORK_PROJECT

FIXTURE_DIR = Path(__file__).resolve().parent


@pytest.mark.datafiles(FIXTURE_DIR)
def test_load_project_from_yaml(datafiles):
    from core.flags import FlagParser
    from core.config.project import Project
    from core.main import parser

    flags = FlagParser(parser, project_dir=str(datafiles))
    project = Project(flags)
    project.load_project_from_yaml()

    assert project.project_dict == EXPECTED_SHEETWORK_PROJECT



import os
from setuptools import find_packages, setup

NAME = "sheetwork"
VERSION = "1.0.0a12"
DESCRIPTION = """
    sheetwork is a command line tool to load sheets from google
    and upload them to snowflake
    """

REQUIRES_PYTHON = ">=3.6.0"

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md")) as f:
    long_description = f.read()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Bastien Boutonnet",
    author_email="bastien.b1@gmail.com",
    install_requires=[
        "requests<2.23.0",
        "gspread==3.3.0",
        "sqlalchemy==1.3.16",
        "cerberus==1.3.2",
        "pandas==1.0.4",
        "pyyaml==5.3.1",
        "snowflake-sqlalchemy==1.2.3",
        "oauth2client==4.1.3",
        "inflection==0.4.0",
        "colorama==0.4.3",
    ],
    python_requires=REQUIRES_PYTHON,
    entry_points={"console_scripts": ["sheetwork = core.main:main"]},
    packages=find_packages(),
    license="MIT",
    url="https://github.com/bastienboutonnet/sheetwork",
    download_url="https://github.com/bastienboutonnet/sheetwork/archive/v1.0.0a12.tar.gz",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
)



import re

import inflection
import pandas


class SheetCleaner:
    def __init__(self, df: pandas.DataFrame, casing: bool = False):
        self.df = df
        self.casing = casing
        assert isinstance(
            self.df, pandas.DataFrame
        ), f"SheetCleaner can only process a pandas.DataFrame. You are feeding a {type(self.df)}."

    def cleanup(self):
        clean_df = self.df.copy(deep=True)

        if self.casing:
            clean_df = self.camel_to_snake(clean_df)
        clean_df = self.columns_cleanups(clean_df)
        clean_df = self.fields_cleanups(clean_df)

        return clean_df

    @staticmethod
    def columns_cleanups(df):

        # clean column names (slashes and spaces to understore), remove trailing whitespace
        df.columns = [re.sub(r"^\d+", "", col) for col in df.columns]
        df.columns = [
            col.replace(" ", "_")
            .replace("/", "_")
            .replace(".", "_")
            .replace("?", "_")
            .replace("__", "_")
            .strip()
            for col in df.columns
        ]
        df.columns = [re.sub(r"^\_+", "", col) for col in df.columns]
        df.columns = [re.sub(r"\_+$", "", col) for col in df.columns]
        df.columns = [re.sub(r"[^\w\s]", "", col) for col in df.columns]

        # remove empty cols
        if "" in df.columns:
            df = df.drop([""], axis=1)

        # make all columns lowercase
        df.columns = map(str.lower, df.columns)
        return df

    @staticmethod
    def fields_cleanups(df):

        # clean trailing spaces in fields
        for col in df.columns:
            if df[col].dtype == "object":
                df[col] = df[col].str.strip()

        return df

    @staticmethod
    def camel_to_snake(df: pandas.DataFrame) -> pandas.DataFrame:
        df.columns = [inflection.underscore(col) for col in df.columns]
        return df
